"""
Provide utilities for concurrent programming.
"""

import asyncio
import multiprocessing
import threading as stdthreading
import time
from abc import ABC, abstractmethod
from asyncio import sleep
from collections.abc import Hashable, MutableMapping
from ctypes import c_longdouble
from math import floor
from multiprocessing.managers import SyncManager
from types import TracebackType
from typing import Self, TypeAlias, TypeVar, Union, final

from typing_extensions import override

from betty.typing import processsafe
from betty.warnings import deprecate

_KeyT = TypeVar("_KeyT")
_ValueT = TypeVar("_ValueT")


class Lock(ABC):
    """
    Provide an asynchronous lock.
    """

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.release()

    @abstractmethod
    async def acquire(self, *, wait: bool = True) -> bool:
        """
        Acquire the lock.
        """

    @abstractmethod
    async def release(self) -> None:
        """
        Release the lock.
        """


Acquirable: TypeAlias = Union[stdthreading.Lock, stdthreading.Semaphore]  # noqa: UP007


async def asynchronize_acquire(acquirable: Acquirable, *, wait: bool = True) -> bool:
    """
    Acquire a synchronous lock or semaphore asynchronously.
    """
    while not acquirable.acquire(blocking=False):
        if not wait:
            return False
        # Sleeping for zero seconds does not actually sleep, but gives the event
        # loop a chance to progress other tasks while we wait for another chance
        # to acquire the acquirable.
        await sleep(0)
    return True


@final
class AsynchronizedLock(Lock):
    """
    Make a synchronous (blocking) lock asynchronous (non-blocking).
    """

    __slots__ = "_lock"

    def __init__(self, lock: stdthreading.Lock):
        self._lock = lock

    @property
    def lock(self) -> stdthreading.Lock:
        """
        The underlying, synchronous lock.
        """
        return self._lock

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        return await asynchronize_acquire(self._lock, wait=wait)

    @override
    async def release(self) -> None:
        self._lock.release()

    @classmethod
    def threading(cls) -> Self:
        """
        Create a new thread-safe, asynchronous lock.
        """
        return cls(stdthreading.Lock())


class Semaphore(ABC):
    """
    An asynchronous semaphore.
    """

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.release()

    @abstractmethod
    async def acquire(self, *, wait: bool = True) -> bool:
        """
        Acquire the semaphore.
        """

    @abstractmethod
    async def release(self, n: int = 1) -> None:
        """
        Release the semaphore.
        """


@final
class AsynchronizedSemaphore(Semaphore):
    """
    Make a synchronous (blocking) semaphore asynchronous (non-blocking).
    """

    __slots__ = "_semaphore"

    def __init__(self, semaphore: stdthreading.Semaphore):
        self._semaphore = semaphore

    @property
    def semaphore(self) -> stdthreading.Semaphore:
        """
        The underlying, synchronous semaphore.
        """
        return self._semaphore

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        return await asynchronize_acquire(self._semaphore, wait=wait)

    @override
    async def release(self, n: int = 1) -> None:
        self._semaphore.release(n)

    @classmethod
    def threading(cls, n: int = 1) -> Self:
        """
        Create a new thread-safe, asynchronous semaphore.
        """
        return cls(stdthreading.Semaphore(n))


@final
@processsafe
class RateLimiter:
    """
    Rate-limit operations.

    This class implements the `Token Bucket algorithm <https://en.wikipedia.org/wiki/Token_bucket>`_.
    """

    def __init__(
        self, maximum: int, period: int = 1, *, manager: SyncManager | None = None
    ):
        manager = ensure_manager(manager)
        self._lock = AsynchronizedLock(manager.Lock())
        self._maximum = maximum
        self._period = period
        self._available = manager.Value(c_longdouble, maximum)
        # A Token Bucket fills as time passes. However, we want callers to be able to start
        # using the limiter immediately, so we 'preload' the first's period's tokens, and
        # set the last added time to the end of the first period. This ensures there is no
        # needless waiting if the number of tokens consumed in total is less than the limit
        # per period.
        self._last_add = manager.Value(c_longdouble, time.monotonic() + self._period)

    def _add_tokens(self):
        now = time.monotonic()
        elapsed = now - self._last_add.value
        added = elapsed * self._maximum
        possibly_available = floor(self._available.value + added)
        if possibly_available > 0:
            self._available.value = min(possibly_available, self._maximum)
            self._last_add.value = now

    async def __aenter__(self) -> None:
        await self.wait()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        return

    async def is_available(self) -> bool:
        """
        Whether an operation may be performed (again).
        """
        async with self._lock:
            self._add_tokens()
            return self._available.value != 0

    async def wait(self) -> None:
        """
        Wait until an operation may be performed (again).
        """
        async with self._lock:
            while self._available.value < 1:
                self._add_tokens()
                if self._available.value < 1:
                    await asyncio.sleep(0)
            self._available.value -= 1


class _Transaction(Lock):
    def __init__(
        self,
        transaction_id: Hashable,
        ledger_lock: Lock,
        ledger: MutableMapping[Hashable, bool],
    ):
        self._transaction_id = transaction_id
        self._ledger_lock = ledger_lock
        self._ledger = ledger

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        if wait:
            while True:
                async with self._ledger_lock:
                    if self._can_acquire():
                        return self._acquire()
                await sleep(0)
        else:
            async with self._ledger_lock:
                if self._can_acquire():
                    return self._acquire()
                return False

    def _can_acquire(self) -> bool:
        try:
            return not self._ledger[self._transaction_id]
        except KeyError:
            self._ledger[self._transaction_id] = False
            return True

    def _acquire(self) -> bool:
        self._ledger[self._transaction_id] = True
        return True

    @override
    async def release(self) -> None:
        self._ledger[self._transaction_id] = False


@processsafe
class Ledger:
    """
    Lazily create locks by keeping a ledger.

    The ledger lock is released once a transaction lock is acquired.
    """

    def __init__(self, ledger_lock: Lock, *, manager: SyncManager | None = None):
        manager = ensure_manager(manager)
        self._ledger_lock = ledger_lock
        self._ledger: MutableMapping[Hashable, bool] = manager.dict()

    def ledger(self, transaction_id: Hashable) -> Lock:
        """
        Ledger a new lock for the given transaction ID.
        """
        return _Transaction(transaction_id, self._ledger_lock, self._ledger)


def ensure_manager(manager: SyncManager | None, *, stacklevel: int = 1) -> SyncManager:
    """
    Ensure that a value is a multiprocessing manager.
    """
    if manager:
        return manager
    deprecate(
        "Not providing a multiprocessing manager is deprecated as of Betty 0.4.10.",
        stacklevel=stacklevel,
    )
    return multiprocessing.Manager()



"""
Provide Jinja2 filters to integrate with Webpack.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from jinja2 import pass_context

from betty.project.extension.webpack.jinja2 import _context_js_entry_points

if TYPE_CHECKING:
    from jinja2.runtime import Context

    from betty.jinja2 import Filters


@pass_context
async def filter_webpack_entry_point_js(
    context: Context, entry_point_name: str
) -> None:
    """
    Add a Webpack entry point's JavaScript files to the current page.
    """
    _context_js_entry_points(context).add(entry_point_name)


FILTERS: Filters = {
    "webpack_entry_point_js": filter_webpack_entry_point_js,
}



"""
Populate ancestries with information from Wikipedia.
"""

from __future__ import annotations

from asyncio import gather
from collections import defaultdict
from contextlib import suppress
from typing import TYPE_CHECKING

from betty.ancestry.file import File
from betty.ancestry.file_reference import FileReference
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.link import HasLinks, Link
from betty.ancestry.place import Place
from betty.concurrent import AsynchronizedLock, Lock
from betty.fetch import FetchError
from betty.functools import filter_suppress
from betty.locale import (
    UNDETERMINED_LOCALE,
    Localey,
    get_data,
    negotiate_locale,
    to_locale,
)
from betty.locale.error import LocaleError
from betty.media_type.media_types import HTML
from betty.typing import internal, threadsafe
from betty.wikipedia import NotAPageError, parse_page_url

if TYPE_CHECKING:
    from collections.abc import Mapping, MutableMapping, MutableSequence, Sequence

    from betty.ancestry import Ancestry
    from betty.locale.localizer import LocalizerRepository
    from betty.wikipedia.client import Client, Image, Summary
    from betty.wikipedia.copyright_notice import WikipediaContributors


@internal
@threadsafe
class Populator:
    """
    Populate an ancestry with information from Wikipedia.
    """

    def __init__(
        self,
        ancestry: Ancestry,
        locales: Sequence[str],
        localizers: LocalizerRepository,
        client: Client,
        copyright_notice: WikipediaContributors,
    ):
        self._ancestry = ancestry
        self._locales = locales
        self._localizers = localizers
        self._client = client
        self._image_files: MutableMapping[Image, File] = {}
        self._image_files_locks: Mapping[Image, Lock] = defaultdict(
            AsynchronizedLock.threading
        )
        self._copyright_notice = copyright_notice

    async def populate(self) -> None:
        """
        Populate the ancestry.
        """
        await gather(
            *(
                self._populate_entity(entity, self._locales)
                for entity in self._ancestry
                if isinstance(entity, HasLinks)
            )
        )

    async def _populate_entity(self, entity: HasLinks, locales: Sequence[str]) -> None:
        populations = [self._populate_has_links(entity, locales)]
        if isinstance(entity, HasFileReferences):
            populations.append(self._populate_has_file_references(entity))
        if isinstance(entity, Place):
            populations.append(self._populate_place(entity))
        await gather(*populations)

    async def _populate_has_links(
        self, has_links: HasLinks, locales: Sequence[str]
    ) -> None:
        summary_links: MutableSequence[tuple[str, str]] = []
        for link in has_links.links:
            try:
                page_language, page_name = parse_page_url(link.url)
            except NotAPageError:
                continue
            else:
                try:
                    get_data(page_language)
                except LocaleError:
                    continue
                else:
                    summary_links.append((page_language, page_name))

            summary = None
            if not link.label:
                with suppress(FetchError):
                    summary = await self._client.get_summary(page_language, page_name)
            await self._populate_link(link, page_language, summary)
        await self._populate_has_links_with_translation(
            has_links, locales, summary_links
        )

    async def _populate_has_links_with_translation(
        self,
        has_links: HasLinks,
        locales: Sequence[str],
        summary_links: MutableSequence[tuple[str, str]],
    ) -> None:
        for page_language, page_name in summary_links:
            page_translations = await self._client.get_translations(
                page_language, page_name
            )
            if len(page_translations) == 0:
                continue
            page_translation_locale_datas: Sequence[Localey] = list(
                filter_suppress(get_data, LocaleError, page_translations.keys())
            )
            for locale in locales:
                if locale == page_language:
                    continue
                added_page_locale_data = negotiate_locale(
                    locale, page_translation_locale_datas
                )
                if added_page_locale_data is None:
                    continue
                added_page_language = to_locale(added_page_locale_data)
                added_page_name = page_translations[added_page_language]
                if (added_page_language, added_page_name) in summary_links:
                    continue
                added_summary = await self._client.get_summary(
                    added_page_language, added_page_name
                )
                if not added_summary:
                    continue
                added_link = Link(added_summary.url)
                await self._populate_link(
                    added_link, added_page_language, added_summary
                )
                has_links.links.append(added_link)
                summary_links.append((added_page_language, added_page_name))
            return

    async def _populate_link(
        self, link: Link, summary_language: str, summary: Summary | None = None
    ) -> None:
        if link.url.startswith("http:"):
            link.url = "https:" + link.url[5:]
        if link.media_type is None:
            link.media_type = HTML
        if link.relationship is None:
            link.relationship = "external"
        if link.locale is UNDETERMINED_LOCALE:
            link.locale = summary_language
        if not link.description:
            # There are valid reasons for links in locales that aren't supported.
            with suppress(ValueError):
                link.description = (
                    await self._localizers.get_negotiated(link.locale)
                )._("Read more on Wikipedia.")
        if summary is not None and not link.label:
            link.label[summary_language] = summary.title

    async def _populate_place(self, place: Place) -> None:
        await self._populate_place_coordinates(place)

    async def _populate_place_coordinates(self, place: Place) -> None:
        await gather(
            *(
                self._populate_place_coordinates_link(place, link)
                for link in place.links
            )
        )

    async def _populate_place_coordinates_link(self, place: Place, link: Link) -> None:
        try:
            page_language, page_name = parse_page_url(link.url)
        except NotAPageError:
            return
        else:
            coordinates = await self._client.get_place_coordinates(
                page_language, page_name
            )
            if coordinates:
                place.coordinates = coordinates

    async def _populate_has_file_references(
        self, has_file_references: HasFileReferences & HasLinks
    ) -> None:
        await gather(
            *(
                self._populate_has_file_references_link(has_file_references, link)
                for link in has_file_references.links
            )
        )

    async def _populate_has_file_references_link(
        self, has_file_references: HasFileReferences & HasLinks, link: Link
    ) -> None:
        try:
            page_language, page_name = parse_page_url(link.url)
        except NotAPageError:
            return
        else:
            image = await self._client.get_image(page_language, page_name)
            if not image:
                return
            await self._image_file_reference(has_file_references, image)

    async def _image_file_reference(
        self, has_file_references: HasFileReferences, image: Image
    ) -> FileReference:
        async with self._image_files_locks[image]:
            try:
                file = self._image_files[image]
            except KeyError:
                links = []
                for locale in self._locales:
                    localizer = await self._localizers.get(locale)
                    links.append(
                        Link(
                            f"{image.wikimedia_commons_url}?uselang={locale}",
                            label=localizer._(
                                "Description, licensing, and image history"
                            ),
                            description=localizer._(
                                "Find out more about this image on Wikimedia Commons."
                            ),
                            locale=locale,
                            media_type=HTML,
                        )
                    )
                file = File(
                    id=f"wikipedia-{image.title}",
                    name=image.name,
                    path=image.path,
                    media_type=image.media_type,
                    links=links,
                    copyright_notice=self._copyright_notice,
                )
                self._image_files[image] = file
                self._ancestry.add(file)
            file_reference = FileReference(has_file_references, file)
            self._ancestry.add(file_reference)
            return file_reference



"""
Interact with the Wikipedia Query API.
"""

from __future__ import annotations

import logging
from contextlib import contextmanager
from dataclasses import dataclass
from json import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast, final
from urllib.parse import quote, urlparse

from geopy import Point

from betty.fetch import Fetcher, FetchError
from betty.locale.localizable import plain
from betty.media_type import MediaType
from betty.typing import internal

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping, MutableMapping

    from betty.concurrent import RateLimiter


@final
@dataclass(frozen=True)
class Summary:
    """
    A Wikipedia page summary.
    """

    locale: str
    name: str
    title: str
    content: str

    @property
    def url(self) -> str:
        """
        The URL to the web page.
        """
        return f"https://{self.locale}.wikipedia.org/wiki/{self.name}"


@final
@dataclass(frozen=True)
class Image:
    """
    An image from Wikimedia Commons.
    """

    path: Path
    media_type: MediaType
    title: str
    wikimedia_commons_url: str
    name: str


RATE_LIMIT = 200


@internal
class Client:
    """
    Fetch information from the Wikipedia Query API.
    """

    def __init__(self, fetcher: Fetcher, rate_limiter: RateLimiter):
        self._fetcher = fetcher
        self._images: MutableMapping[str, Image | None] = {}
        self._rate_limiter = rate_limiter

    @contextmanager
    def _catch_exceptions(self) -> Iterator[None]:
        try:
            yield
        except FetchError as error:
            logging.getLogger(__name__).warning(str(error))

    async def _fetch_json(self, url: str, *selectors: str | int) -> Any:
        async with self._rate_limiter:
            response = await self._fetcher.fetch(url)
        try:
            data = response.json
        except JSONDecodeError as error:
            raise FetchError(
                plain(f"Invalid JSON returned by {url}: {error}")
            ) from error

        try:
            for selector in selectors:
                data = data[selector]
        except (LookupError, TypeError) as error:
            raise FetchError(
                plain(
                    f"Could not successfully parse the JSON format returned by {url}: {error}"
                )
            ) from error
        return data

    async def _get_query_api_data(self, url: str) -> Mapping[str, Any]:
        return cast(
            "Mapping[str, Any]", await self._fetch_json(url, "query", "pages", 0)
        )

    async def _get_page_query_api_data(
        self, page_language: str, page_name: str
    ) -> Mapping[str, Any]:
        return await self._get_query_api_data(
            f"https://{page_language}.wikipedia.org/w/api.php?action=query&titles={quote(page_name)}&prop=langlinks|pageimages|coordinates&lllimit=500&piprop=name&pilicense=free&pilimit=1&coprimary=primary&format=json&formatversion=2"
        )

    async def get_translations(
        self, page_language: str, page_name: str
    ) -> Mapping[str, str]:
        """
        Get the available translations for a page.
        """
        try:
            api_data = await self._get_page_query_api_data(page_language, page_name)
        except FetchError as error:
            logger = logging.getLogger(__name__)
            logger.warning(str(error))
            return {}
        try:
            translations_data = api_data["langlinks"]
        except LookupError:
            # There may not be any translations.
            return {}
        return {
            translation_data["lang"]: translation_data["title"]
            for translation_data in translations_data
        }

    async def get_summary(self, page_language: str, page_name: str) -> Summary | None:
        """
        Get a summary for a page.
        """
        with self._catch_exceptions():
            url = f"https://{page_language}.wikipedia.org/api/rest_v1/page/summary/{page_name}"
            api_data = await self._fetch_json(url)
            try:
                return Summary(
                    page_language,
                    page_name,
                    api_data["titles"]["normalized"],
                    (
                        api_data["extract_html"]
                        if "extract_html" in api_data
                        else api_data["extract"]
                    ),
                )
            except LookupError as error:
                raise FetchError(
                    plain(
                        f"Could not successfully parse the JSON content returned by {url}: {error}"
                    )
                ) from error

    async def get_image(self, page_language: str, page_name: str) -> Image | None:
        """
        Get an image for a page.
        """
        with self._catch_exceptions():
            api_data = await self._get_page_query_api_data(page_language, page_name)
            try:
                page_image_name = api_data["pageimage"]
            except LookupError:
                # There may not be any images.
                return None

            if page_image_name in self._images:
                return self._images[page_image_name]

            url = f"https://en.wikipedia.org/w/api.php?action=query&prop=imageinfo&titles=File:{quote(page_image_name)}&iiprop=url|mime|canonicaltitle&format=json&formatversion=2"
            image_info_api_data = await self._get_query_api_data(url)

            try:
                image_info = image_info_api_data["imageinfo"][0]
            except LookupError as error:
                raise FetchError(
                    plain(
                        f"Could not successfully parse the JSON content returned by {url}: {error}"
                    )
                ) from error
            async with self._rate_limiter:
                image_path = await self._fetcher.fetch_file(image_info["url"])
            return Image(
                image_path,
                MediaType(image_info["mime"]),
                # Strip "File:" or any translated equivalent from the beginning of the image's title.
                image_info["canonicaltitle"][
                    image_info["canonicaltitle"].index(":") + 1 :
                ],
                image_info["descriptionurl"],
                Path(urlparse(image_info["url"]).path).name,
            )

    async def get_place_coordinates(
        self, page_language: str, page_name: str
    ) -> Point | None:
        """
        Get the coordinates for a page that is a place.
        """
        with self._catch_exceptions():
            api_data = await self._get_page_query_api_data(page_language, page_name)
            try:
                coordinates = api_data["coordinates"][0]
            except LookupError:
                # There may not be any coordinates.
                return None
            try:
                if coordinates["globe"] != "earth":
                    return None
                return Point(coordinates["lat"], coordinates["lon"])
            except LookupError as error:
                raise FetchError(
                    plain(f"Could not successfully parse the JSON content: {error}")
                ) from error



"""
Fetch information from Wikipedia.
"""

from __future__ import annotations

import re
from typing import cast


class NotAPageError(ValueError):
    """
    Raised when a URL does not point to a Wikipedia page.
    """


_PAGE_URL_PATTERN = re.compile(r"^https?://([a-z]+)\.wikipedia\.org/wiki/([^/?#]+).*$")


def parse_page_url(url: str) -> tuple[str, str]:
    """
    Parse the URL for a Wikipedia page.

    :return: A 2-tuple with the page language and the page name.
    """
    match = _PAGE_URL_PATTERN.fullmatch(url)
    if match is None:
        raise NotAPageError
    return cast(tuple[str, str], match.groups())



"""
Data types for people's presences at events.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from typing_extensions import override

from betty.locale.localizable import Localizable, _, ngettext
from betty.model import Entity
from betty.model.association import BidirectionalToOne, ToOneAssociate
from betty.plugin import ShorthandPluginBase
from betty.privacy import (
    HasPrivacy,
    Privacy,
    is_public,
    merge_secondary_privacies,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from betty.ancestry.event import Event
    from betty.ancestry.person import Person
    from betty.ancestry.presence_role import PresenceRole
    from betty.json.linked_data import JsonLdObject
    from betty.mutability import Mutable
    from betty.project import Project
    from betty.serde.dump import Dump, DumpMapping


@final
class Presence(ShorthandPluginBase, HasPrivacy, Entity):
    """
    The presence of a :py:class:`betty.ancestry.person.Person` at an :py:class:`betty.ancestry.event.Event`.
    """

    _plugin_id = "presence"
    _plugin_label = _("Presence")

    #: The person whose presence is described.
    person = BidirectionalToOne["Presence", "Person"](
        "betty.ancestry.presence:Presence",
        "person",
        "betty.ancestry.person:Person",
        "presences",
        title="Person",
    )
    #: The event the person was present at.
    event = BidirectionalToOne["Presence", "Event"](
        "betty.ancestry.presence:Presence",
        "event",
        "betty.ancestry.event:Event",
        "presences",
        title="Event",
    )
    #: The role the person performed at the event.
    role: PresenceRole

    def __init__(
        self,
        person: ToOneAssociate[Person],
        role: PresenceRole,
        event: ToOneAssociate[Event],
        *,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(None, privacy=privacy, public=public, private=private)
        self.person = person
        self.role = role
        self.event = event

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        return (self.role,)

    @override
    @classmethod
    def plugin_label_plural(cls) -> Localizable:
        return _("Presences")

    @override
    @classmethod
    def plugin_label_count(cls, count: int) -> Localizable:
        return ngettext("{count} presence", "{count} presences", count).format(
            count=str(count)
        )

    @override
    @property
    def label(self) -> Localizable:
        return _("Presence of {person} at {event}").format(
            person=self.person.label,
            event=self.event.label,
        )

    @override
    def _get_effective_privacy(self) -> Privacy:
        return merge_secondary_privacies(
            super()._get_effective_privacy(),
            self.person,
            self.event,
        )

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "role", await project.presence_role_repository.plugin_id_schema, False
        )
        return schema

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)
        if is_public(self):
            dump["role"] = self.role.plugin_id()
        return dump
